/*
** Actions handling.
**
** $Id: action.c,v 1.13 2001/02/25 18:04:10 kilobug Exp $
*/

#include "ant.h"
#include "ant_private.h"

/* Do absolutely nothing */
error_t	ant_do_nothing(ant_t *ant, const void *param)
{
  return ERROR_SUCCESS;
}

/* Check if an ant can go to target cell */
error_t ant_check_target(ant_t *ant, const coord_t *where)
{
  if (where == NULL)
    return ERROR_PARAM;
  if (ant_dist(where, &(ant->pos)) == 0)
    return ERROR_NONE;
  if ((ant->world->rules->jump_size > 0) &&
      (ant_dist(where, &(ant->pos)) > ant->world->rules->jump_size))
    return ERROR_FORBIDDEN;
  if (!ant_cell_empty(ant->world, where))
    return ERROR_FORBIDDEN;
  return ERROR_SUCCESS;
}

/* Performs a move */
void	ant_do_move_to(ant_t *ant, const coord_t *where)
{
  ant_delete_item(ant->world, &(ant->pos), cont_ant, ant->id);
  if (ant_find_item(ant->world, where, cont_exit) != NULL)
    {
      ant->world->nb_die++;
      ant_kill_ant(ant);
      return;
    }

  ant_add_item(ant->world, where, cont_ant, ant->id);

  ant->pos.x = where->x;
  ant->pos.y = where->y;
}

/* Move ant to "where" */
error_t	ant_move_to(ant_t *ant, const coord_t *where)
{
  error_t err;

  err = ant_check_target(ant, where);
  if (err != ERROR_SUCCESS)
    return err;
  ant_do_move_to(ant, where);
  return ERROR_NONE;
}

/* Push wood to "where" */
error_t	ant_push_to(ant_t *ant, const coord_t *where)
{
  error_t err;
  ant_t *p;
  content_t *wood;
  int param;
  GSList *push = NULL, *l;

  /* Check the target cell */
  err = ant_check_target(ant, where);
  if (err != ERROR_SUCCESS)
    return err;
  if (ant_find_item(ant->world, where, cont_wood) != NULL)
    return ERROR_FORBIDDEN;

 /* Find the wood */
  wood = ant_find_item(ant->world, &(ant->pos), cont_wood);
  if (wood == NULL)
    return ERROR_PARAM;

  /* Build the pushing list */
  for (l = ant->world->ants; l != NULL; l = g_slist_next(l))
    {
      p = l->data;
      g_assert(p != NULL);
      if ((ant_dist(&(ant->pos), &(p->pos)) == 0) &&
	  (p->curact == ant->curact) &&
	  (ant_dist(p->curact_data, ant->curact_data) == 0))
	push = g_slist_prepend(push, p);
    }

  /* Check is enough ants are pushing */
  if ((g_slist_length(push) < wood->param) || (wood->param <= 0))
    {
      g_slist_free(push);
      return ERROR_TOO_HEAVY;
    }

  /* Ok, move all */
  param = wood->param;
  ant_delete_item(ant->world, &(ant->pos), wood->type, wood->param);
  if (ant_find_item(ant->world, where, cont_home) != NULL)
    ant->world->wood_home += param;
  else
    ant_add_item(ant->world, where, cont_wood, param);
  ant->world->wood_total = ant->world->wood_map + ant->world->wood_home;
  for (l = push; l != NULL; l = g_slist_next(l))
    ant_do_move_to(l->data, where);

  g_slist_free(push);
  return ERROR_NONE;
}

/* Push or move */
error_t	ant_push_move_to(ant_t *ant, const coord_t *where)
{
  if (ant_push_to(ant, where) != ERROR_NONE)
    return ant_move_to(ant, where);
  else
    return ERROR_NONE;
}

/* Perform an action */
error_t	ant_action(ant_t *ant, int action, void *param)
{
  actions_t *l;
  error_t res;

  /* Check params and find action */
  if (ant == NULL)
    return ERROR_PARAM;
  for (l = ant->world->actions; l != NULL; l = l->next)    
    if (l->id == action)
      break;
  if (l == NULL)
    return ERROR_PARAM;

  /* Set the action value */
  ant_enter_critical();
  ant->curact = action;
  ant->curact_data = param;
  ant_leave_critical();

  /* Wait */
  ant_apply_delay(&(ant->world->rules->live), ant->world);
  if (ant->die)
    ant_kill_ant(ant);
  
  /* Performs action */
  ant_enter_critical();
  ant->world->nb_live++;
  ant_stats_have_changed(ant->world);
  res = l->func(ant, param);
  ant->curact = ANT_ACT_NONE;
  ant->curact_data = NULL;
  ant_leave_critical();

  return res;
}

/* Add an action to the list of known actions */
error_t	ant_add_action(world_t *w, int action, ant_action_func_t *func)
{
  actions_t *act;

  if (w == NULL || func == NULL)
    return (ERROR_PARAM);
  act = g_malloc(sizeof (*act));
  act->id = action;
  act->func = func;
  act->next = w->actions;  
  w->actions = act;
  return (ERROR_NONE);
}

