/*
** Contains general types.
**
** $Id: ant.h,v 1.32 2001/02/25 18:04:10 kilobug Exp $
*/

#ifndef __ANT_H__
#define __ANT_H__

#include "ant_general.h"

/*******************
 ** General types **
 *******************/

typedef char error_t;
#define ERROR_SUCCESS	1	/* In some cases this is needed */
#define ERROR_NONE	0
#define ERROR_PARAM	-1	/* Invalid parameter: out of range, ... */
#define ERROR_MALLOC	-2	/* Out of memory */
#define ERROR_PARSE	-3	/* Error in config file */
#define ERROR_FORBIDDEN	-10	/* Action forbidden */
#define ERROR_TOO_HEAVY	-11	/* Wood too heavy */
#define ERROR_OTHER	-42	/* Other error */

typedef struct coord_s
{
  int x, y;
} coord_t;

/***********
 ** Rules **
 ***********/

typedef struct actions_s actions_t;
typedef struct world_s world_t;

typedef struct delay_s
{
  int min, max;
} delay_t;

typedef struct rules_s
{
  coord_t size;            /* Size of the map */
  int nb_ant_max;	   /* Number of maximal ants at the same time */
  int nb_ant_total;	   /* Number of maximal ants to create */
  int max_by_cell;         /* Number of ants per cell */
  int jump_size;	   /* Maximal movement size, in norm 1 */
  delay_t create;          /* Delay between two ants creation */
  delay_t live;		   /* Delay between two actions for an ant */
  coord_t entry;	   /* Coords of the default entry point */
  char *conf_name;	   /* Name of the configuration file */
  char *map_name;	   /* Name of the map file */
  char *plugin_name;       /* Name of the plugin file or NULL in static mode */
  int cell_size;	   /* Size of a cell in GFX display */
  int paving;		   /* Size of the pavement in GFX display */
  char *font_name;         /* X name of the font for display */
  void *data;		   /* Data for external program if needed */
} rules_t;

/**********
 ** Ants **
 **********/

typedef struct ant_s
{
  int id;                  /* Unique number specific of this ant */
  coord_t pos;		   /* Current position */
  void *data;		   /* Private data for the ant */ 
  world_t *world;	   /* A pointer to the main world */

  /* This should normally only be used internally */
  gboolean die;            /* TRUE: ant will die at the next cycle */
  int curact;		   /* Current action */
  void *curact_data;	   /* Current action data */
} ant_t;

typedef error_t ant_callback_t(ant_t *ant);

/***********
 ** World **
 ***********/

typedef enum { 
  cont_wall, cont_ant, cont_exit, cont_wood, cont_entry, cont_home 
} content_type_t;

typedef struct content_s
{
  content_type_t type;     /* Type of the element */
  int param;		   /* Element parameter (size of wood, ant id) */
  struct content_s *next;  /* Pointer the next element (linked list) */
} content_t;

struct world_s
{
  int nb_ant;		   /* Number of ants currently living */
  coord_t size;		   /* Size of the world */
  actions_t *actions;      /* Linked-list of known actions */
  ant_callback_t *live;    /* Callback function for each cycle */
  ant_callback_t *born;	   /* Callback initializing function */
  ant_callback_t *exit;	   /* Callback destructor function */
  rules_t *rules;	   /* The rules */
  content_t **map;	   /* Array of lists of elems present on the world */  
  GSList *entry;	   /* List of entry points */

  /* Stats */
  int nb_born;             /* Number of ants created since the last reset */
  int nb_die;		   /* Number of ants exited since the last reset */
  int nb_live;             /* Number of call to live since the last reset */
  int wood_home;	   /* Units of wood at home */
  int wood_map;		   /* Units of wood currently on the map */
  int wood_total;	   /* Sum of the two above */
  
  GSList *ants;            /* List of all ants */
  void *gfx_data;	   /* Data for the graphical interface */
  void *data;		   /* Data for external program */
};

#define MAP(W, X, Y) ((W).map[X + Y * (W).size.x])

/*************
 ** Actions **
 *************/

#define ANT_ACT_NONE		0
#define ANT_ACT_MOVE		1
#define ANT_ACT_PUSH		2	
#define ANT_ACT_PUSH_MOVE	3  /*  PUSH or MOVE */	

typedef error_t ant_action_func_t(ant_t *ant, void *param);

struct actions_s
{
  int id;
  ant_action_func_t *func;
  actions_t *next;
};

/*************************
 ** Functions prototype **
 *************************/

#include "ant_world.h"
#include "ant_engine.h"
#include "ant_gtk_public.h"

#endif

