/*
** Ants handling: creating them and making them live.
**
** $Id: ants.c,v 1.24 2001/02/24 23:41:47 kilobug Exp $
*/

#include "ant.h"
#include "ant_private.h"

GMutex *ants_mutex = NULL;

/* Get an unique number - thread safe */
int ant_get_uniq_id()
{
  static int id = 0;
  static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
  int res;

  g_static_mutex_lock(&mutex);
  res = id++;  
  g_static_mutex_unlock(&mutex);
  return res;
}

/* Wait */
void ant_apply_delay(const delay_t *delay, world_t *w)
{
  int d;
  
  ant_action_begin(w);
  if (delay->max == 0)
    return;
  d = delay->max - delay->min;
  if (d != 0)
    d = random() % d;
  d += delay->min;
  if (d != 0)
    usleep(d * 100);
  ant_action_end(w);
}

/******************
** Creating ants **
*******************/

/* Create a new ant */
error_t ant_create_ant(world_t *w)
{
  ant_t *the_ant;
  pthread_t thread;

  if (w == NULL)
    return ERROR_PARAM;
  if (w->rules == NULL)
    return ERROR_PARAM;  
  ant_apply_delay(&(w->rules->create), w);
  if (((w->nb_ant < w->rules->nb_ant_max) || (w->rules->nb_ant_max < 0)) &&
      ((w->nb_born < w->rules->nb_ant_total) || (w->rules->nb_ant_total < 0)))
    {
      ant_enter_critical();
      if (ant_cell_empty(w, &(w->rules->entry)))
	{
	  w->nb_ant++;
	  ALLOC(the_ant);
	  the_ant->world = w;
	  w->ants = g_slist_prepend(w->ants, the_ant);
	  pthread_create(&thread, NULL, (void *) ant_new_ant, the_ant);
	  pthread_detach(thread);
	}
      else
	ant_leave_critical();
    }
  return ERROR_NONE;
}

/* Callback for the pthread_create function */
void *ant_new_ant(ant_t *ant)
{
  error_t err;
  coord_t *c;
  int i;

  i = g_slist_length(ant->world->entry);
  i = random() % i;
  c = g_slist_nth_data(ant->world->entry, i);
  ant->id = ant_get_uniq_id();
  ant->data = NULL;
  ant->pos.x = c->x;
  ant->pos.y = c->y;
  ant->die = FALSE;

  ant_add_item(ant->world, &(ant->pos), cont_ant, ant->id);
  err = ant->world->born(ant);
  ant->world->nb_born++;
  ant_leave_critical();

  if (err != ERROR_NONE)
    fprintf(stderr, "Warning: 'born' returned %d\n", err);
  while (42)
    {
      err = ant->world->live(ant);
      if (err != ERROR_NONE)
	fprintf(stderr, "Warning: 'live' returned %d\n", err);
    }
  return NULL;
}

/********************
** Destroying ants **
*********************/

/* Kill the current ant: destroy the thread, free memory, ... */
void ant_kill_ant(ant_t *ant)
{
  error_t err;

  err = ant->world->exit(ant);
  if (err != ERROR_NONE)
    fprintf(stderr, "Warning: 'exit' returned %d\n", err);
  ant_delete_item(ant->world, &(ant->pos), cont_ant, ant->id);
  ant->world->nb_ant--;
  ant->world->ants = g_slist_remove(ant->world->ants, ant);
  g_free(ant);
  g_mutex_unlock(ants_mutex);
  pthread_exit(NULL);
}

/* Send the KILL signal to all ants */
void ant_kill_all_ants(world_t *w)
{
  GSList *p;

  g_mutex_lock(ants_mutex);
  for (p = w->ants; p != NULL; p = g_slist_next(p))
    ((ant_t *) (p->data))->die = TRUE;
  g_mutex_unlock(ants_mutex);
}

