/*
** Get a value from config file.
**
** $Id: parser_value.c,v 1.5 2001/02/21 20:42:46 kilobug Exp $
*/

#include "ant.h"
#include "ant_private.h"
#include "ant_parser.h"

/* Integer */
error_t ant_parse_int(GScanner *scan, int *res)
{
  GTokenType type;

  type = g_scanner_get_next_token(scan);
  if (type == G_TOKEN_INT)
    {
      *res = g_scanner_cur_value(scan).v_int;
      return ERROR_NONE;
    }
  if (type == G_TOKEN_SYMBOL)
    {
      *res = *((int *) g_scanner_cur_value(scan).v_symbol);
      return ERROR_NONE;
    }
  g_scanner_unexp_token(scan, G_TOKEN_INT, NULL, NULL, NULL, NULL,
			TRUE);
  return ERROR_PARSE;
}

/* String */
error_t ant_parse_string(GScanner *scan, char **res)
{
  GTokenType type;

  SAFE_FREE(*res);
  type = g_scanner_get_next_token(scan);
  if (type == G_TOKEN_STRING)
    {
      *res = strdup(g_scanner_cur_value(scan).v_string);
      return ERROR_NONE;
    }
  if (type == G_TOKEN_SYMBOL)
    {
      *res = strdup(((char *) g_scanner_cur_value(scan).v_symbol));
      return ERROR_NONE;
    }
  g_scanner_unexp_token(scan, G_TOKEN_STRING, NULL, NULL, NULL, NULL,
			TRUE);
  return ERROR_PARSE;
}

/* Coordinates */
error_t ant_parse_coord_paren(GScanner *scan, coord_t *res)
{
  error_t err;
  int x, y;
  
  err = ant_parse_int(scan, &x);
  if (err != ERROR_NONE)
    return err;
  PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_COMMA);
  err = ant_parse_int(scan, &y);
  if (err != ERROR_NONE)
    return err;
  PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_RIGHT_PAREN);
  res->x = x;
  res->y = y;
  return ERROR_NONE;
}

error_t ant_parse_coord_curly(GScanner *scan, coord_t *res)
{
  error_t err;
  int x = -666, y = -666;
  int cx, tmp, cy;
  char *s;
  GTokenType type;
  
  while ((type = g_scanner_get_next_token(scan)) != G_TOKEN_RIGHT_CURLY)
    {
      PARSER_CHECK_TYPE(type, G_TOKEN_IDENTIFIER);
      cx = cy = FALSE;
      s = g_scanner_cur_value(scan).v_identifier;
      if ((!strcmp(s, "x")) || (!strcmp(s, "width")))
	cx = TRUE;
      else
	if ((!strcmp(s, "y")) || (!strcmp(s, "height")))
	  cy = TRUE;
	else
	  if (!strcmp(s, "both"))
	    cx = cy = TRUE;
	  else
	    {
	      g_scanner_error(scan, "invalid keyword: %s", s);
	      return ERROR_PARSE;
	    }
      PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_EQUAL_SIGN);
      err = ant_parse_int(scan, &tmp);
      if (err != ERROR_NONE)
	return err;      
      if (cx)
	x = tmp;
      if (cy)
	y = tmp;
      while (g_scanner_peek_next_token(scan) == G_TOKEN_COMMA)
	g_scanner_get_next_token(scan);
    }
  if (x == -666)
    g_scanner_warn(scan, "value `x' not specified - using default");
  else
    res->x = x;
  if (y == -666)
    g_scanner_warn(scan, "value `y' not specified - using default");
  else
    res->y = y;
  return ERROR_NONE;
}

error_t ant_parse_coord(GScanner *scan, coord_t *res)
{
  GTokenType type;
  coord_t *p;

  type = g_scanner_get_next_token(scan);
  if (type == G_TOKEN_SYMBOL)
    {
      p = g_scanner_cur_value(scan).v_symbol;
      res->x = p->x;
      res->y = p->y;
      return ERROR_NONE;
    }
  if (scan->scope_id == SCOPE_POS)
    g_scanner_set_scope(scan, SCOPE_POS_ELEM);
  else
    g_scanner_set_scope(scan, SCOPE_NONE);
  if (type == G_TOKEN_LEFT_PAREN)
    return ant_parse_coord_paren(scan, res);
  if (type == G_TOKEN_LEFT_CURLY)
    return ant_parse_coord_curly(scan, res);
  g_scanner_error(scan, "expecting valid coordinates");
  return ERROR_PARSE;
}

/* Delay */
error_t ant_parse_delay_paren(GScanner *scan, delay_t *res)
{
  error_t err;
  int min, max;
  
  err = ant_parse_int(scan, &min);
  if (err != ERROR_NONE)
    return err;
  PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_COMMA);
  err = ant_parse_int(scan, &max);
  if (err != ERROR_NONE)
    return err;
  PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_RIGHT_PAREN);
  res->min = min;
  res->max = max;
  return ERROR_NONE;
}

error_t ant_parse_delay_curly(GScanner *scan, delay_t *res)
{
  error_t err;
  int min = -666, max = -666;
  int cmin, tmp, cmax;
  char *s;
  GTokenType type;
  
  while ((type = g_scanner_get_next_token(scan)) != G_TOKEN_RIGHT_CURLY)
    {
      PARSER_CHECK_TYPE(type, G_TOKEN_IDENTIFIER);
      cmin = cmax = FALSE;
      s = g_scanner_cur_value(scan).v_identifier;
      if ((!strcmp(s, "min")) || (!strcmp(s, "minimum")))
	cmin = TRUE;
      else
	if ((!strcmp(s, "max")) || (!strcmp(s, "maximum")))
	  cmax = TRUE;
	else
	  if ((!strcmp(s, "both")) || (!strcmp(s, "value")))
	    cmin = cmax = TRUE;
	  else
	    {
	      g_scanner_error(scan, "invalid keyword: %s", s);
	      return ERROR_PARSE;
	    }
      PARSER_CHECK_TYPE(g_scanner_get_next_token(scan), G_TOKEN_EQUAL_SIGN);
      err = ant_parse_int(scan, &tmp);
      if (err != ERROR_NONE)
	return err;      
      if (cmin)
	min = tmp;
      if (cmax)
	max = tmp;
      while (g_scanner_peek_next_token(scan) == G_TOKEN_COMMA)
	g_scanner_get_next_token(scan);
    }
  if (min == -666)
    g_scanner_warn(scan, "value `min' not specified - using default");
  else
    res->min = min;
  if (max == -666)
    g_scanner_warn(scan, "value `max' not specified - using default");
  else
    res->max = max;
  return ERROR_NONE;
}

error_t ant_parse_delay(GScanner *scan, delay_t *res)
{
  GTokenType type;
  delay_t *p;

  type = g_scanner_get_next_token(scan);
  if (type == G_TOKEN_SYMBOL)
    {
      p = g_scanner_cur_value(scan).v_symbol;
      res->min = p->min;
      res->max = p->max;
      return ERROR_NONE;
    }
  if (type == G_TOKEN_INT)
    {
      res->max = res->min = g_scanner_cur_value(scan).v_int;
      return ERROR_NONE;
    }
  g_scanner_set_scope(scan, SCOPE_NONE);
  if (type == G_TOKEN_LEFT_PAREN)
    return ant_parse_delay_paren(scan, res);
  if (type == G_TOKEN_LEFT_CURLY)
    return ant_parse_delay_curly(scan, res);
  g_scanner_error(scan, "expecting valid delay value");
  return ERROR_PARSE;
}

/* We've found an identifier - get the name & value */
error_t ant_parser_get_value(GScanner *scan, rules_t *rules, 
			     const char *section)
{
  char *what = NULL;
  int i;
  guint scope;
  error_t err = ERROR_OTHER;;

  what = strdup(g_scanner_cur_value(scan).v_identifier);
  if (g_scanner_get_next_token(scan) != G_TOKEN_EQUAL_SIGN)
    {
      g_scanner_unexp_token(scan, G_TOKEN_EQUAL_SIGN, NULL, NULL, NULL, NULL,
			    TRUE);
      SAFE_FREE(what);
      return ERROR_PARSE;
    }
  for (i = 0; ant_config_file[i].name != NULL; i++)
    if (!s_strcasecmp(section, ant_config_file[i].section) && 
	(!strcmp(what, ant_config_file[i].name)))
      break;
  if (ant_config_file[i].name == NULL)
    {
      g_scanner_error(scan, "unkwon identifier `%s' in section %s", what, 
		      section);
      SAFE_FREE(what);
      return ERROR_PARSE;
    }
  scope = g_scanner_set_scope(scan, ant_config_file[i].scope_id);
  switch (ant_config_file[i].type)
    {
    case pt_int:
      err = ant_parse_int(scan, 
			  &(G_STRUCT_MEMBER(int, rules, 
					    ant_config_file[i].offset)));
      break;
    case pt_string: 
      err = ant_parse_string(scan, 
			     &(G_STRUCT_MEMBER(char *, rules, 
					       ant_config_file[i].offset)));
      break;
    case pt_coord:
      err = ant_parse_coord(scan, 
			    &(G_STRUCT_MEMBER(coord_t, rules, 
					      ant_config_file[i].offset)));
      break;
    case pt_delay:
      err = ant_parse_delay(scan, 
			    &(G_STRUCT_MEMBER(delay_t, rules, 
					      ant_config_file[i].offset)));
      break;
    case pt_none:
      err = ERROR_NONE;
      break;
    }
  g_scanner_set_scope(scan, scope);
  SAFE_FREE(what);
  return err;
}
