/*
** Contains the world functions.
**
** $Id: world.c,v 1.23 2001/02/25 18:04:10 kilobug Exp $
*/

#include "ant.h"
#include "ant_private.h"

/* Returns TRUE if an ant can go to this cell */
int ant_cell_empty(const world_t *w, const coord_t *where)
{
  int nb_ants = 0;
  content_t *p;

  if ((where->x < 0) || (where->x >= w->size.x))
    return FALSE;
  if ((where->y < 0) || (where->y >= w->size.y))
    return FALSE;
  for (p = MAP(*w, where->x, where->y); p != NULL; p = p->next)
    {
      if (p->type == cont_wall)
	return FALSE;
      if (p->type == cont_ant)	
	if ((++nb_ants >= w->rules->max_by_cell) && 
	    (w->rules->max_by_cell >= 0))
	  return FALSE;
    }
  return TRUE;
}

/* Create an empty world from the rules */
world_t	*ant_create_world(rules_t *rules)
{
  world_t *ret;
  int i, j, x, y;

  if (rules == NULL)
    return NULL;
  ret = g_malloc(sizeof (*ret));

  ret->nb_live = ret->nb_die = ret->nb_born = ret->nb_ant = 0;
  ret->wood_total = ret->wood_home = ret->wood_map = 0;
  x = ret->size.x = rules->size.x;
  y = ret->size.y = rules->size.y;
  ret->rules = rules;
  ret->data = NULL;
  ret->ants = NULL;
  ret->entry = NULL;

  ret->map = g_malloc(sizeof (*(ret->map)) * x * y);
  for (i = 0; i < x; i++)
    for (j = 0; j < y; j++)
      MAP(*ret, i, j) = NULL;

  ret->actions = NULL;
  ant_add_action(ret, ANT_ACT_NONE, (void *) ant_do_nothing);
  ant_add_action(ret, ANT_ACT_MOVE, (void *) ant_move_to);
  ant_add_action(ret, ANT_ACT_PUSH, (void *) ant_push_to);
  ant_add_action(ret, ANT_ACT_PUSH_MOVE, (void *) ant_push_move_to);
  return (ret);
}

/* Same as add_item, but don't call the ant_cell_has_changed callback */
void ant_do_add_item(world_t *w, const coord_t *where, content_type_t type, 
		     int par)
{
  coord_t *nw;
  content_t **p, *q;

  switch (type)
    {
    case cont_wood:
      if (par > 0)
	w->wood_map += par;
      w->wood_total = w->wood_map + w->wood_home;
      break;
    case cont_entry:
      ALLOC(nw);
      nw->x = where->x;
      nw->y = where->y;
      w->entry = g_slist_prepend(w->entry, nw);
      break;
    case cont_home:
    case cont_exit:
    case cont_wall:
    case cont_ant:
      break;
    }
  p = &(MAP(*w, where->x, where->y));
  ALLOC(q);
  q->type = type;
  q->param = par;
  q->next = *p;
  *p = q;  
}

/* Add an item at the following coords */
void ant_add_item(world_t *w, const coord_t *where, 
		  content_type_t type, int par)
{
  ant_do_add_item(w, where, type, par);
  ant_cell_has_changed(w, where->x, where->y);
}

/* Delete an item from the world */
void ant_delete_item(world_t *w, const coord_t *where, 
		     content_type_t type, int par)
{
  GSList *l;
  content_t **p, *q;

  p = &(MAP(*w, where->x, where->y));
  while (*p != NULL)
    {
      if (((*p)->type == type) && ((*p)->param == par))
	{
	  switch (type)
	    {
	    case cont_wood:
	      if (par > 0)
		w->wood_map -= par;
	      break;
	    case cont_entry:
	      for (l = w->entry; l != NULL; l = g_slist_next(l))
		if (ant_dist(where, l->data) == 0)
		  break;
	      if (l != NULL)
		w->entry = g_slist_remove(w->entry, l->data);
	      break;
	    case cont_home:
	    case cont_exit:
	    case cont_wall:
	    case cont_ant:
	      break;
	    }
	  q = (*p)->next;
	  g_free(*p);
	  *p = q;
	  break;
	}
      p = &((*p)->next);
    }
  w->wood_total = w->wood_map + w->wood_home;
  ant_cell_has_changed(w, where->x, where->y);
}

/* Find an item */
content_t *ant_find_item(const world_t *w, const coord_t *where, 
			 content_type_t type)
{
  content_t *p;

  p = MAP(*w, where->x, where->y);
  for (; p != NULL; p = p->next)
    if (p->type == type)
      return p;
  return NULL;
}

/* Returns the distance between two points (in norm 1) */
int ant_dist(const coord_t *a, const coord_t *b)
{
  return (abs(a->x - b->x) + abs(a->y - b->y));
}
